// module txtpdf.cpp for UNIX/Dos/Win 32 bit shell        
// Generate a PDF file from a ASCII text file      
// The ASCII file may contain simple tags similar to HTML
// (but not necessarily compatible)

// Private Copyright Dr.E.Huckert, Goldbach(Germany) 5-97   
// Mail address: huckert@compuserve.com                     

// Revisions:
// 01.04.97 Created Version 1.0
// 25.09.97 Version 1.4: classes revised and simplified
// 08.09.02 Version 1.5: created: error (BT-ET) fixed in WriteHeader() 

// Conventions and commands
//   <FONT fontno size leading>
//   <NEED nolines>
//   <PAGE>  inserts a page break
//   <P>     inserts a paragraph break

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/stat.h>

#ifdef UNIX
#include <unistd.h>
#define setmode(x,y)
#define CREAT_MASK  0755
#else
#include <io.h>
#define CREAT_MASK 0xff        /* only for DOS/Windows */
#endif

#define LOW_HIGH      0
#define HIGH_LOW      1
#define MAXLINE      80     // wrap position for header line 2
#define MLINES       80     // max. number of lines per page
//
// A4 Format = 210 mm * 297 mm == (8.3*72)  (11.7*72) in inch/72
// must be modified for American paper sizes
#define PAGE_YDIM   842
#define PAGE_XDIM   597
#define TOP_MARGIN   35
#define LEFT_MARGIN  25
#define LEADING      15
#define FONTSIZE     12
#define FONTNO       1
//
// The following values can be used to place an image on the page
// They are for European A4 paper size
#define CM1          550.0 
#define CM2          0.0
#define CM3          0.0
#define CM4          700.0
#define CM5          20.0 
#define CM6          40.0 

// global variables
int     logging     = 0;

// a PDF object: the objects form a simple chain
class PDFOBJ
{
public:
  int    no;            // number of the object
  short  written;       // flag: 1 if object has been written to the file
  short  various;       // various information (font number etc.)
  long   objpos;        // Write position in the file
  char   dsc[20];       // Description/Type of object
  PDFOBJ *next;         // pointer to the next object
public:
  // the constructor
  PDFOBJ(int  numb, 
         int  vari,
         long pos, 
         char *name)
  {
    this->no     = numb; 
    this->various= vari;
    this->objpos = pos;
    this->next   = NULL;
    if (strlen(name) < 20)
      strcpy(this->dsc, name);
  };
};

class PDF
{
private:
  int  ofd;
  int  no_objs;
  int  root_objno;
  int  info_objno;
  int  pages_objno;
  int  linewidth;
  int  leading;
  int  pagecount;
  long xrefpos;
  long writepos;
  long imglength;
  PDFOBJ *g_first_obj;       // pointer to the first PDFOBJ 
  PDFOBJ *g_last_obj; 
  char  page_mode[80];
  char  viewer_prefs[120];
  int   rotation;
  int   bookmarks;
  int   adjust;
  int   my_byte_order;
  int   act_fontno;
  int   act_fontsize;  
  short font[20];
  //
  // member variables related to a single page
  int   page_xdim;
  int   page_ydim;
  int   top_margin;
  int   left_margin;
  float cm1;         // for matrix operator cm
  float cm2;
  float cm3;
  float cm4;
  float cm5;
  float cm6;
#ifdef IMAGE
  int   tiff_xdim;
  int   tiff_ydim;
#endif
  //
  // member variables for the text lines on a single page
  int  no_text_lines;
  int  textpar[MLINES][4];
  char *textptr[MLINES];
  //
  // Variables for text parsing and tag detection
  int  fontno;
  int  paragraph;
  int  pagebreak;
  int  actionflag;
  int  fontsize;
  int  current_leading;
  int  font_change;
  int  current_ypos;
  //
  // local (page related) chain of PDF objects
  PDFOBJ *p_first_obj;       // pointer to the first PDFOBJ 
  PDFOBJ *p_last_obj; 
  //
  int   openflag;            // 1 if desctructor must close file ofd
  int   global_error;
  //
public:
  PDF(char *filename, char *subject);
  PDF(int fd, char *subject);
  ~PDF();  
  int GetGlobalError()
  { return global_error; };
  void BasicInit();
  int  CreateEssentialObjects();
  int  ReadInitialisations();
  int  StartObject(int wr_flag);
  int  StartObject(char *type, int wr_flag);
  int  StartObject(char *type, int vari, int wr_flag);
  int  CloseObject();
  int  AddLine(char *line);
  int  AddDict(char *line);
  int  StartDict();
  int  CloseDict();
  int  GetObjectCount()
  { return no_objs; };
  int  GetRootObjno()
  {  return root_objno; };
  int  GetPagesObjno()
  {  return pages_objno; };
  int  GetPageCount()
  {  return pagecount; };
  void SetPageCount(int pc)
  {  pagecount = pc; };
  int  GetLineWidth()
  {  return linewidth; };
  void SetLineWidth(int lw)
  {  linewidth = lw; };
  void SetLeading(int ld)
  {  leading = ld; };
  int GetLeftMargin()
  {  return left_margin; };
  void SetPageDimensions(int xd, int yd)
  {  page_xdim = xd; page_ydim = yd; };
  int GetPageXDim()
  {  return page_xdim; };
  int GetPageYDim()
  {  return page_ydim; };
  int GetLeading()
  {  return leading; };
  int MakeAnnotation(int x1, int y1, 
                     int x2, int y2, char *txt);
  int MakeLinkAnnotation(int link_to, 
                         int x1, int y1, 
                         int x2, int y2);
  int MakeExtLinkAnnotation(char *filename,
                            int x1, int y1, 
                            int x2, int y2);
  int MakeInfo(char *subject, char *title, char *author, 
               char *creator, char *date);
  int  MakeFont(int fontno, char *fontname);
  int  MakePagesObject();
  int  WritePagesObject(int mode);
  int  WritePageResources();
  int  StartPage(int pages_obj, char *subtype);
  int  ClosePage();
  int  ClosePage(int parent, char *ptype);
  int  MakePage(int pages_obj, char *subtype);
  int  RegisterPageObject(int objno, char *ptype);
  void DeletePageObjects();
  void DeleteGlobalObjects();
  int  InitTextContent();
  int  CollectTextContent(int fontno, 
                         int fontsize,
                         int xpos, int ypos, 
                         char *txt);
  int  WriteTextContent();
  int  ReleaseTextContent();
  int  WriteHeader(char *head1, char *head2);
  int  WriteFooter(char *txt, int nopage, int pag_style);
  char *MakeHeader2(int ypos, int lead, char *head2);
  char *PrepareTextLine(char *txt);
  int  GetObjectByType(char *typ, int mode);
  PDFOBJ *GetObjectStart()
  {  return g_first_obj; };
  //
  int GetMyByteOrder();
  int GetByteOrder()
  { return my_byte_order; };
  char *GetDate();
  int InsertTextFromFile(char *filename,
                         char *header1,
                         char *header2,
                         char *footer);
};

// Generate an entry in the local trace file
void put_trace(char *msg)
{
static long count = 0L;
FILE *fd;
  if (count++ == 0L)
    fd = fopen("txtpdf.txt", "w");
  else
    fd = fopen("txtpdf.txt", "a");
  if (fd != NULL)
  {
    fprintf(fd, "%s", msg);
    fclose(fd);
  }
}

// Determine the byte order of the machine on which we are
// running
// In text mode normally not necessary
int PDF::GetMyByteOrder()
{
char *caddr;
static short testword;
  testword = 0x0040;
  caddr = (char *)&testword;
  if (*caddr == 0)
    return HIGH_LOW;
  else
    return LOW_HIGH;
}   // end GetMyByteOrder()

// Reads from file txtpdf.ini
// Entries there have the form variable=wert 
// Be cautious: the syntax is less flexible than in Windows!!!
int PDF::ReadInitialisations()
{
char hbuf[128];
char *p;
FILE *ifil;
  ifil = fopen("txtpdf.ini", "r");
  if (ifil == NULL)
    return -1;
  while (1)
  {
    if (fgets(hbuf, sizeof(hbuf), ifil) == NULL)
      break;
    if (strlen(hbuf) <= 3) continue;
    if (hbuf[0] == ';')
      // Kommentar
      continue;
    p = strchr(hbuf, '=');
    if (p == NULL) continue;
    *p = 0;
    if (strcmp(hbuf, "logging") == 0)
      ::logging = atoi(p+1);
    if (strcmp(hbuf, "page_xdim") == 0)
      this->page_xdim = atoi(p+1);
    if (strcmp(hbuf, "page_ydim") == 0)
      this->page_ydim = atoi(p+1);
    if (strcmp(hbuf, "leading") == 0)
      this->leading = atoi(p+1);
    if (strcmp(hbuf, "cm1") == 0)
      this->cm1 = atof(p+1);
    if (strcmp(hbuf, "cm2") == 0)
      this->cm2 = atof(p+1);
    if (strcmp(hbuf, "cm5") == 0)
      this->cm5 = atof(p+1);
    if (strcmp(hbuf, "cm6") == 0)
      this->cm6 = atof(p+1);
    if (strcmp(hbuf, "cm4") == 0)
      this->cm4 = atof(p+1);
    if (strcmp(hbuf, "cm3") == 0)
      this->cm3 = atof(p+1);
    if (strcmp(hbuf, "pagemode") == 0)
      // muss als kompletter Befehl erscheinen
      strcpy(this->page_mode, p + 1);
    if (strcmp(hbuf, "viewer") == 0)
      // muss als kompletter Befehl erscheinen
      strcpy(this->viewer_prefs, p + 1);
  }
  fclose(ifil);
  return 0;
}   // end PDF::ReadInitialisations()

void PDF::BasicInit()
{
int n;
  no_objs     = 0;
  ofd         = 0;
  info_objno  = 0;
  root_objno  = 0;
  pages_objno = 0;
  xrefpos     = 0L;
  writepos    = 0L;
  leading     = LEADING;         // line leading
  linewidth   = 2;               // line width in 1/72 inch
  pagecount   = 0;
  page_xdim     = PAGE_XDIM;
  page_ydim     = PAGE_YDIM;
  top_margin    = TOP_MARGIN;
  left_margin   = LEFT_MARGIN;
  cm1           = CM1;
  cm2           = CM2;
  cm3           = CM3;
  cm4           = CM4; 
  cm5           = CM5;
  cm6           = CM6;
  strcpy(page_mode, "");
  strcpy(viewer_prefs, "");
  rotation      = 0;            // page rotation
  bookmarks     = 0;
  adjust        = 0;
  current_ypos  = page_ydim;
  act_fontsize  = FONTSIZE;
  act_fontno    = FONTNO;
  my_byte_order = this->GetMyByteOrder();
#ifdef IMAGE
  tiff_xdim     = 0;      // X dim. of actual TIFF image
  tiff_ydim     = 0;      // Y dim. of actual TIFF image
#endif
  // initialize global objects chain  
  g_first_obj   = NULL;
  g_last_obj    = NULL;
  // initialize local objects chain
  p_first_obj   = NULL;
  p_last_obj    = NULL;
  //
  fontno        = 1;
  paragraph     = 0;
  pagebreak     = 0;
  actionflag    = 0;
  fontsize      = leading - 1;
  current_leading= leading;
  font_change    = 0;
  //
  openflag      = 0;
  global_error  = 0;
}   // end PDF::BasicInit()

int PDF::CreateEssentialObjects()
{
  this->AddLine("%PDF-1.1\n");  
  font[0] = -1;
  font[1] = this->MakeFont(1, "Helvetica");
  font[2] = this->MakeFont(2, "Helvetica-Bold");
  font[3] = this->MakeFont(3, "Helvetica-Oblique");
  font[4] = this->MakeFont(4, "Helvetica-BoldOblique");
  //
  font[5] = this->MakeFont(5, "Times-Roman");
  font[6] = this->MakeFont(6, "Times-Bold");
  font[7] = this->MakeFont(7, "Times-Italic");
  font[8] = this->MakeFont(8, "Times-BoldItalic");
  //
  font[9]  = this->MakeFont(9, "Courier");
  font[10] = this->MakeFont(10, "Courier-Bold");
  font[11] = this->MakeFont(11, "Courier-Oblique");
  font[12] = this->MakeFont(12, "Courier-BoldOblique");
  //
  font[13] = this->MakeFont(13, "Symbol");
  font[14] = this->MakeFont(14, "ZapfDingbats");
  //
  this->StartObject("Procset", 1);
  this->AddLine("[/PDF /Text /ImageB]\n");
  this->CloseObject();
  this->StartObject("/Pages", 0);
  return 0;
}   // end PDF::CreateEssentialObjects()

// constructor
PDF::PDF(char *filename,
         char *subject)
{
  this->BasicInit();
  this->openflag = 1;
  this->ReadInitialisations();
  ofd = creat(filename, CREAT_MASK);
  if (ofd < 0)
  {
    put_trace("Error: cannot create output file\n");
    global_error++;
  }
  else
  {
    setmode(ofd, O_BINARY);
    this->CreateEssentialObjects();
    //
    this->MakeInfo(subject,
                   "",
                   "Dr.E.Huckert, Goldbach (Germany)", 
                   "Programm txtpdf V1.5 by E.Huckert", 
                   this->GetDate());
  }
  InitTextContent();
}  // end PDF::PDF()

// constructor with an (open) file handle
PDF::PDF(int fd, char *subject)
{
  this->BasicInit();
  this->ReadInitialisations();
  this->ofd = fd;
  if (this->ofd < 0)
  {
    put_trace("Error: Invalid Output File\n");
    exit(2);
  }
  setmode(this->ofd, O_BINARY);
  this->CreateEssentialObjects();
  this->MakeInfo(subject,
                 "",
                 "Dr.E.Huckert, Goldbach (Germany)", 
                 "Programm txtpdf", 
                 this->GetDate());
  InitTextContent();
}  // end PDF::PDF()

// Destructor
// Writes the /Pages object
// Generates a trailer with the cross reference of objects
PDF::~PDF()
{
int    n, root_obj;
char   hbuf[128];
PDFOBJ *pobj;
  if (logging)
    put_trace("PDF::~PDF()\n");
  if (ofd != 0)
  {
    // Generate the /Kids array in the /Pages object
    n = this->WritePagesObject(2);
    if (n < 0)
    {
      put_trace("/Pages Creation Error\n");
      global_error++;
    }
    //
    // Generate a /Root object with a reference to the /Pages object
    root_obj = this->StartObject("/Root", 1);
    sprintf(hbuf, "/Type /Catalog\n/Pages %d 0 R\n", 
            this->GetPagesObjno());
    if (strlen(page_mode) > 0)     
    {
      // visualisation mode
      strcat(hbuf, page_mode);
    }
    if (strlen(viewer_prefs) > 0)    
    {
      // visualisation preferences
      strcat(hbuf, viewer_prefs);
    }
    this->AddDict(hbuf);
    n = this->CloseObject();
    if (n < 0)
    {
      put_trace("Write Error\n");
      global_error++;
    }
    // 
    // generate the cross reference
    this->xrefpos = this->writepos;
    sprintf(hbuf, "xref\n0 %d\n", this->no_objs + 1);
    this->AddLine(hbuf);
    /* a line must have exactly 20 chars incl. the linefeed */
    this->AddLine("0000000000 65535 f \n");
    pobj = this->GetObjectStart();
    while (pobj != NULL)
    {
      sprintf(hbuf, "%010ld 00000 n \n", pobj->objpos);
      this->AddLine(hbuf);
      pobj = pobj->next;
    }
    //
    // generate the trailer
    this->AddLine("trailer\n<<\n");
    sprintf(hbuf, "/Size %d\n", this->no_objs + 1);
    this->AddLine(hbuf);
    if (this->root_objno != 0)
    {
      sprintf(hbuf, "/Root %d 0 R\n", root_objno);
      this->AddLine(hbuf);
    }
    if (this->info_objno != 0)
    {
      sprintf(hbuf, "/Info %d 0 R\n", info_objno);
      this->AddLine(hbuf);
    }
    this->CloseDict();
    sprintf(hbuf,"startxref\n%ld\n", this->xrefpos);
    this->AddLine(hbuf);
    write(this->ofd,"%%EOF\n", 6);
    close(ofd);
  }
  // delete the collected global objects
  this->DeleteGlobalObjects();
  if (this->openflag)
    close(this->ofd);
  ReleaseTextContent();
}   // end PDF::~PDF()

int PDF::GetObjectByType(char *ty, int local)
{
PDFOBJ *pobj;
  if (local == 0)
    // search in global chain
    pobj = this->GetObjectStart();
  else
   // search in local chain
   pobj = this->p_first_obj;
  while (pobj != NULL)
  {
    if (strcmp(pobj->dsc, ty) == 0)
      return pobj->no;
    pobj = pobj->next;
  }
  return 0;
}   // end PDF::GetObjectByType()

// Starts a new object
// Adds the object to the chain of objects of type PDFOBJ
// Returns the number of the new object
// if write_flag <> 0: the object is written on the output file
int PDF::StartObject(char *type, 
                     int write_flag,
                     int vari)
{
PDFOBJ *pobj;
char    hbuf[128];
int     nb;
  pobj = new PDFOBJ(this->no_objs + 1, vari, this->writepos, type);
  if (pobj == NULL)
  {
    put_trace( "-- Mem. Allocation Error, StartObject() --\n");
    global_error++;
    return -1;
  }
  if (this->GetObjectStart() == NULL)
    this->g_first_obj = pobj;
  else
    (this->g_last_obj)->next = pobj;
  this->g_last_obj = pobj;
  this->no_objs++;
  if (write_flag)
  {
    sprintf(hbuf, "%d 0 obj\n", this->no_objs);
    nb = this->AddLine(hbuf);
  }
  if (strcmp(type, "/Root") == 0)
    this->root_objno = this->no_objs;
  if (strcmp(type, "/Info") == 0)
    this->info_objno = this->no_objs;
  if (strcmp(type, "/Pages") == 0)
    this->pages_objno = this->no_objs;
  pobj->written = write_flag;
  return pobj->no;
}   // end PDF::StartObject()

// Starts a new object
// Adds the object to the chain of objects of type PDFOBJ
// Returns the number of the new object
// if write_flag <> 0: the object is written on the output file
int PDF::StartObject(char *type, 
                     int write_flag)
{
int no;
  no = this->StartObject(type, write_flag, 0);
  return no;
}   // end PDF::StartObject()

int PDF::StartObject(int write_flag)
{
int no;
  no = this->StartObject("", write_flag);
  return no;
}   // end PDF::StartObject()

// write a sequence of bytes in a PDF object
int PDF::AddLine(char *str)
{
int nb;
  nb = write(this->ofd, str, strlen(str));
  this->writepos += nb;
  return nb;
}   // end PDF::AddLine()

int PDF::StartDict()
{
int nb;
  nb = this->AddLine("<<\n");
  return nb;
}   // end PDF::StartDict()

int PDF::CloseDict()
{
int nb;
  nb = this->AddLine(">>\n");
  return nb;
}   // end PDF::CloseDict()

// Generate a PDF dictionary between "<<" and ">>"
int PDF::AddDict(char *str)
{
int nb;
  this->StartDict();
  nb = this->AddLine(str);
  this->CloseDict();
  return nb;
}   // end PDF::AddDict()

int PDF::CloseObject()
{
int nb;
  nb = this->AddLine("endobj\n");
  return nb;
}   // end PDF::CloseObject()

// Generate an annotation object
int PDF::MakeAnnotation(
                        int x1, int y1, 
                        int x2, int y2,
                        char *txt)
{
char hbuf[256];
int nb, no;
  no = this->StartObject("/Annot", 1);
  if (no < 0)
    return -1;
  this->StartDict();
  this->AddLine("/Type /Annot /Subtype /Text\n");
  sprintf(hbuf, "/Rect [%d %d %d %d]\n", x1, y1, x2, y2);
  this->AddLine(hbuf);
  this->AddLine("/Contents ");
  if (txt != NULL)
  {
    sprintf(hbuf, "(%s)\n", txt);
    this->AddLine(hbuf);
  }
  this->CloseDict();
  nb = this->CloseObject();
  return no;
}   // end PDF::MakeAnnotation()

// Generates a link annotation object
int PDF::MakeLinkAnnotation(int link_to,
                            int x1, int y1, int x2, int y2)
{
char hbuf[256];
int  nb, no;
  no = this->StartObject("/Link", 1);
  if (no < 0)
    return -1;
  this->StartDict();
  this->AddLine("/Type /Annot /Subtype /Link\n");
  sprintf(hbuf, "/Rect [%d %d %d %d]\n", x1, y1, x2, y2);
  this->AddLine(hbuf);
  this->AddLine("/C [1 0 0]\n");    // red rectangle as link symbol
  this->AddLine("/Border [0 0 2]\n"); 
  sprintf(hbuf, "/Dest [%d 0 R /XYZ 0 0 0]\n", link_to);
  this->AddLine(hbuf);
  this->CloseDict();
  nb = this->CloseObject();
  return no;
}   // end PDF::MakeLinkAnnotation()

// Generate an external link annotation (link to another document)
// Returns: the number of the generated link object
int PDF::MakeExtLinkAnnotation(char *ext_fn,
                               int x1, int y1, 
                               int x2, int y2)
{
char hbuf[256];
int  nb, no;
  if (logging)
    put_trace("MakeExtLinkAnnotation()\n");
  no = this->StartObject("/ExtLink", 1);
  if (no < 0)
    return -1;
  this->StartDict();
  this->AddLine("/Type /Annot /Subtype /Link\n");
  sprintf(hbuf, "/Rect [%d %d %d %d]\n", x1, y1, x2, y2);
  this->AddLine(hbuf);
  this->AddLine("/C [1 0 0]\n");    // red rectangle as link symbol
  this->AddLine("/Border [0 0 2]\n"); 
  sprintf(hbuf, 
          "/A << /Type /Action /S /GoToR /D [ 0 /FitBH %d] /F (%s) >>\n",
          this->page_ydim, ext_fn);
  this->AddLine(hbuf);
  this->CloseDict();
  nb = this->CloseObject();
  return no;
}   // end PDF::MakeExtLinkAnnotation()

// Generate an info object
int PDF::MakeInfo(char *subject,
                  char *title,
                  char *author, 
                  char *creator, 
                  char *date)
{
int  nb, no;
char hbuf[256];
  no = this->StartObject("/Info", 1);
  if (no < 0)
    return -1;
  this->StartDict();
  if (subject != NULL)
  {
    sprintf(hbuf, "/Subject (%s)\n", subject);
    this->AddLine(hbuf);
  }
  if (title != NULL)
  {
    sprintf(hbuf, "/Title (%s)\n", title);
    this->AddLine(hbuf);
  }
  if (creator != NULL)
  {
    sprintf(hbuf, "/Creator (%s)\n", creator);
    this->AddLine(hbuf);
  }
  if (author != NULL)
  {
    sprintf(hbuf, "/Author (%s)\n", author);
    this->AddLine(hbuf);
  }
  if (date != NULL)
  {
    sprintf(hbuf, "/CreationDate (%s)\n", date);
    this->AddLine(hbuf);
  }
  this->CloseDict();
  nb = this->CloseObject();
  return no;
}   // end PDF::MakeInfo()

// Initialize the text content member variables
int  PDF::InitTextContent()
{ 
int n;
  if (logging)
    put_trace("PDF::InitTextContent()\n");
  no_text_lines = 0; 
  for (n=0; n < MLINES; n++)
    this->textptr[n] = NULL;
  return 0;
}  // end PDF::InitTextContent()

// Collect text Lines for a text object
// !! The line of text must have been allocated on the heap (via new...)!!!
// We keep here the pointers, i.e. we make no copy
int PDF::CollectTextContent(int fontno,
                            int fontsize,
                            int xpos, 
                            int ypos,
                            char *txt)
{
int  m;
  if (logging)
    put_trace("PDF::CollectTextContent()\n");
  m = this->no_text_lines;
  if (m < MLINES)
  {
    this->textptr[m] = txt;
    this->textpar[m][0] = fontno;
    this->textpar[m][1] = fontsize;
    this->textpar[m][2] = xpos;
    this->textpar[m][3] = ypos;
    this->no_text_lines += 1;
  }
  else
    return -1;
  return this->no_text_lines;
}   // end PDF::CollectTextContent()

// Write the collected text lines to the text object
// The allocated lines are not freed here
// Returns: the number of the created object
int PDF::WriteTextContent()
{
int  n, nb, no;
char hbuf[256];
char cmds[128];
  if (logging)
    put_trace("PDF::WriteTextContent()\n");
  // Laenge der Textzeilen ermitteln (vermeidet indirektes Objekt)
  nb = 0;
  for (n=0; n < this->no_text_lines; n++)
  {
    if (this->textptr[n] != NULL)
    {
      sprintf(hbuf, "/F%d %d Tf\n1 0 0 1 %d %d Tm (%s) Tj\n", 
              this->textpar[n][0],
              this->textpar[n][1],
              this->textpar[n][2],
              this->textpar[n][3],
              this->textptr[n]);
      nb += strlen(hbuf);
    }
  }
  nb = nb + 3 + 3;   // for "ET\n" and "BT\n"
  //
  no = this->StartObject("Text", 1);
  if (no < 0)
    return -1;
  sprintf(hbuf, "<< /Length %d >>\nstream\n", nb);
  this->AddLine(hbuf);
  this->AddLine("BT\n");
  for (n=0; n < this->no_text_lines; n++)
  {
    sprintf(hbuf, "/F%d %d Tf\n1 0 0 1 %d %d Tm (%s) Tj\n", 
            this->textpar[n][0],
            this->textpar[n][1],
            this->textpar[n][2],
            this->textpar[n][3],
            this->textptr[n]);
    this->AddLine(hbuf);
  }
  nb = this->AddLine("ET\n");
  nb = this->AddLine("endstream\n");
  nb = this->CloseObject();
  return no;
}   // end PDF::WriteTextContent()

// Release the text content, initialize the text content buffer
// Frees the allocated text strings (should not be on the stack!)
// Returns: 0 
int PDF::ReleaseTextContent()
{
int  n;
  if (logging)
    put_trace("PDF::ReleaseTextContent()\n");
  for (n=0; n < this->no_text_lines; n++)
  {
    if (this->textptr[n] != NULL)
      delete this->textptr[n];
  }
  this->InitTextContent();
  return 0;
}   // end PDF::ReleaseTextContent()

// Generate the lower part of the header block
// Make several lines if necessary 
// Returns an allocated buffer that must be freed outside!
char *PDF::MakeHeader2(int ypos,
                       int leading,
                       char *par_kopf2)   // 09.08.96
{
int  break_flag, ll, m, n;
char ch, lch;
char *tbuf2;
char *tbuf1;
char line[MAXLINE + 30];
char hbuf[MAXLINE * 2];
  if (logging)
    put_trace("PDF::MakeHeader2()\n");
  //
  // Generates Escape sequences for '(' and ')'
  ll = (2 * strlen(par_kopf2)) + 1;
  tbuf1 = new char[ll];
  if (tbuf1 == NULL)
  {
    put_trace("Memory Allocation Error\n");
    global_error++;
    return NULL;
  }
  m   = 0;
  lch = 0;
  for (n=0; par_kopf2[n] != 0; n++)
  {
    if (m > (ll - 4))
      break;
    if ((par_kopf2[n] == '(') && (lch != '\\'))
    {
      tbuf1[m++] = '\\';
      tbuf1[m++] = '(';
      lch = '(';
      continue;
    }
    if ((par_kopf2[n] == ')') && (lch != '\\'))
    {
      tbuf1[m++] = '\\';
      tbuf1[m++] = ')';
      lch = ')';
      continue;
    }
    tbuf1[m++] = par_kopf2[n];
    lch        = par_kopf2[n];
  }   // end for n...
  tbuf1[m] = 0;
  //
  ll    = (3 * strlen(par_kopf2) + 100);  
  tbuf2 = new char[ll];
  if (tbuf2 == NULL)
    return NULL;
  tbuf2[0] = 0;
  //
  // break the buffer into lines
  m = 0;
  for (n=0; tbuf1[n] != 0; n++)
  {
    ch = tbuf1[n] & 0xff;
    line[m++] = ch;
    break_flag = 0;
    if (m >= (sizeof(line) - 2))  break_flag = 1;
    //if ((m > MAXLINE) && (ch <= ' ')) break_flag = 1;
    if (m > MAXLINE)
    {
      if (ch <= ' ') break_flag = 1;
      if (strchr("-!:.?/", ch) != NULL) break_flag = 1;
    }
    if (break_flag)
    {
      // this is the wrap position
      line[m] = 0;
      sprintf(hbuf, "BT\n/F%d %d Tf\n1 0 0 1 %d %d Tm (%s) Tj\nET\n",
              FONTNO, FONTSIZE, this->GetLeftMargin(), ypos, line);
      strcat(tbuf2, hbuf);
      m = 0;
      ypos -= leading;
    }
  }
  if (m > 0)
  {
    line[m] = 0;
    sprintf(hbuf, "BT\n/F%d %d Tf\n1 0 0 1 %d %d Tm (%s) Tj\nET\n",
            FONTNO, FONTSIZE, this->GetLeftMargin(), ypos, line);
    strcat(tbuf2, hbuf);
    ypos -= leading;
    m = 0;
  }
  delete tbuf1;
  return tbuf2;
}   // end PDF::MakeHeader2()

// Write an header object consisting of the header text followed
// by a horizontal line
// Assumes that the text head1 has been run thru PrepareTextLine()
// Returns: the number of the created object
// Note: comsumes vertical space in the order of 4 * this->leading
int PDF::WriteHeader(char *head1,
                     char *head2)
{
int header_obj, nb, ypl, ypos;
char hbuf[128];
char tline1[512];
char hline[128];
char *ph2;
  if (logging)
    put_trace("PDF::WriteHeader()\n");
  header_obj = this->StartObject("Text", 1);
  if (header_obj < 0)
    return -1;
  ypos = this->page_ydim - (2 * this->leading);
  sprintf(tline1, "BT\n/F%d %d Tf\n1 0 0 1 %d %d Tm (%s) Tj\nET\n", 
          FONTNO, FONTSIZE, this->left_margin, ypos, head1);
  ypos -= (this->leading / 2);  
  sprintf(hline, "%d w\n%d %d m\n%d %d l S\n",
          this->GetLineWidth(), this->left_margin, ypos,
          this->page_xdim - (2 * this->left_margin), ypos);
  ypos -= this->leading;
  ph2  = this->MakeHeader2(ypos, this->leading, head2);
  sprintf(hbuf, "<< /Length %d >>\nstream\n", 
          strlen(tline1) + strlen(ph2) + strlen(hline));
  this->AddLine(hbuf);
  this->AddLine(tline1);
  this->AddLine(hline);
  this->AddLine(ph2);
  nb = this->AddLine("endstream\n");
  nb = this->CloseObject();
  delete ph2;
  return header_obj;
}   // end PDF::WriteHeader()

// Writes a footer line, containing a page number if desired
// Returns: the number of the created object 
//        : < 0 if error
int PDF::WriteFooter(char *txt,
                     int nopage,    // if >=0 : page numbers are shown
                     int pag_style) // 0=arabic, 1=roman
{
int  nb, no, ypos;
char hbuf[128];
char hline[128];
char tbuf[256];
char pagina[20];
// definitions for max 20 roman style page numbers
static char *roman[] =
{
  "I",
  "II",
  "III",
  "IV",
  "V",
  "VI",
  "VII",
  "VIII",
  "IX",
  "X",
  "XI",
  "XII",
  "XIII",
  "XIV",
  "XV",
  "XVI",
  "XVII",
  "XVIII",
  "XIX",
  "XX"
};
  if (strlen(txt) == 0)
    // no footer text: suppress the horiz. line
    strcpy(hline, "");
  else
  {
    //
    // generate a horizontal separator line
    ypos = (2 * this->leading) + (this->leading / 3);
    sprintf(hline, "%d w\n%d %d m\n%d %d l S\n",
          linewidth, 
          this->left_margin, 
          ypos, 
          this->GetPageXDim() - (2 * this->GetLeftMargin()), 
          ypos);
  }
  //
  // generate the text
  pagina[0] = 0;
  if (nopage >= 0)
  {
    if (pag_style == 0)
      // Arabic digits
      sprintf(pagina, "%d", nopage);
    else
    {
      // roman digits
      if (nopage > 20) nopage = 20;
      strcpy(pagina, roman[nopage-1]);
    }
  }
  if (logging)
  {
    sprintf(hbuf, "PDF::WriteFooter(): pagina=%s\n", pagina);
    put_trace(hbuf);
  }
  sprintf(tbuf, 
         "/F%d %d Tf\n1 0 0 1 %d %d Tm (%s) Tj 1 0 0 1 %d %d Tm (%s) Tj\n", 
         FONTNO, 
         FONTSIZE, 
         this->left_margin, 
         (int)(1.5 * this->leading),
         txt, 
         this->GetPageXDim() - (int)((2.5 * this->GetLeftMargin())), 
         (int)(1.5 * this->leading), 
         pagina);
  //
  no = this->StartObject("Footer", 1);
  if (no < 0)
    return -1;
  nb = strlen(hline)+ strlen(tbuf) + 6;   //  + 6 fuer "ET\n" und "BT\n"
  //nb = strlen(tbuf) + 6;   //  + 6 fuer "ET\n" und "BT\n"
  sprintf(hbuf, "<< /Length %d >>\nstream\n", nb);
  this->AddLine(hbuf);
  this->AddLine(hline);
  this->AddLine("BT\n");
  this->AddLine(tbuf);
  this->AddLine("ET\n");
  this->AddLine("endstream\n");
  nb = this->CloseObject();
  if (nb < 0)
    return -1;
  return no;
}   // end PDF::WriteFooter()

// Makes a font object
// Returns: the egenerated object number
int PDF::MakeFont(int fontno, char *fontname)
{
int nb, no;
char hbuf[128];
  no = this->StartObject("/Font", 1, fontno);
  if (no < 0)
    return -1;
  this->StartDict();
  this->AddLine("/Type /Font\n");
  this->AddLine("/Subtype /Type1\n");
  sprintf(hbuf, "/Name /F%d\n", fontno);
  this->AddLine(hbuf);
  sprintf(hbuf, "/BaseFont /%s\n", fontname);
  this->AddLine(hbuf);
  this->AddLine("/Encoding /WinAnsiEncoding\n");
  this->CloseDict();
  nb = this->CloseObject();
  if (nb < 0)
    return -1;
  return no;
}   // end PDF::MakeFont()

// Generate a /Pages object
// Write the kids (=pages) into an array of type "/Kids"
// Returns: the number of the /Pages object created
int PDF::MakePagesObject()
{
char *kids;
char xbuf[20];
char *hbuf;
int  n, no, no_kids, lkids;
PDFOBJ *pobj;
  if (logging)
    put_trace("PDF::MakePagesObject()\n");
  no = this->StartObject("/Pages", 1);
  //
  // Count the /Page objects
  no_kids = 0;
  pobj = this->GetObjectStart();
  while (pobj != NULL)
  {
    if (strncmp(pobj->dsc, "/Page ", strlen("/Page ")) == 0)
      no_kids++;
    pobj = pobj->next;
  }
  lkids = 20 + (no_kids * 15);
  kids  = new char[lkids];
  if (kids == NULL)
  {
    put_trace("-- Mem.Allocation Error for /Kids 1 --\n");
    global_error++;
    return -1;
  }
  //
  // Generate the /Kids array
  strcpy(kids, "/Kids [");
  pobj = this->GetObjectStart();
  while (pobj != NULL)
  {
    if (strncmp(pobj->dsc, "/Page ", strlen("/Page ")) == 0)
    {
      sprintf(xbuf, "%d 0 R ", pobj->no);
      if ((strlen(xbuf) + strlen(kids)) < lkids)
        strcat(kids, xbuf);
    }
    pobj = pobj->next;
  }
  if ((strlen(xbuf) + strlen("]")) < lkids)
    strcat(kids, "]");
  hbuf = new char[200 + strlen(kids)];
  if (hbuf == NULL)
  {
    put_trace("-- Mem.Allocation Error for /Kids 2 --\n");
    global_error++;
    return -2;
  }
  sprintf(hbuf, "/Type /Pages\n/Count %d\n%s\n",
                this->GetPageCount(), kids);
  this->AddDict(hbuf);
  delete kids;
  delete hbuf;
  n = this->CloseObject();
  if (n < 0)
  {
    put_trace("Write Error\n");
    global_error++;
    return -1;
  }
  return no;
}   // end PDF::MakePagesObject()

// Write a /Pages object
// Assumes that an object of type "/Pages" has been previously 
// allocated in the chain of objects
// Write the kids (=pages) into an array of type "/Kids"
int PDF::WritePagesObject(int mode) // 1=normal order, 2=text pages first
{
char *kids;
char xbuf[40];
char *hbuf;
int  n, no, no_kids, lkids;
PDFOBJ *pobj;
  if (logging)
    put_trace("PDF::WritePagesObject()\n");
  pobj = this->GetObjectStart();
  while (pobj != NULL)
  {
    if (strcmp(pobj->dsc, "/Pages") == 0)
      break;
    pobj = pobj->next;
  }
  if (pobj == NULL)
  {
     // object not identified
     put_trace("Error: /Pages object not found\n");
     global_error++;
     return -1;
  }

  pobj->objpos = writepos;
  no           = pobj->no;
  sprintf(xbuf, "%d 0 obj\n", no);
  n = this->AddLine(xbuf);
  if (n < 0)
  {
    put_trace("Write error\n");
    global_error++;
    return -2;
  }
  pobj->written = 1;

  // Generate the /Kids array
  no_kids = 0;
  pobj = this->GetObjectStart();
  while (pobj != NULL)
  {
    if (strncmp(pobj->dsc, "/Page ", strlen("/Page ")) == 0)
      no_kids++;
    pobj = pobj->next;
  }
  lkids = 50 + (no_kids * 15);
  kids  = new char[lkids];
  if (kids == NULL)
  {
    put_trace("-- Mem.Allocation Error for /Kids 1 --\n");
    global_error++;
    return -1;
  }
  strcpy(kids, "/Kids [");
  if (mode < 2)
  {
    pobj = this->GetObjectStart();
    while (pobj != NULL)
    {
      if (strncmp(pobj->dsc, "/Page ", strlen("/Page ")) == 0)
      {
        sprintf(xbuf, "%d 0 R ", pobj->no);
        if ((strlen(xbuf) + strlen(kids)) < lkids)
          strcat(kids, xbuf);
      }
      pobj = pobj->next;
    }
  }
  else
  {
    // insert text pages first, then image pages
    pobj = this->GetObjectStart();
    while (pobj != NULL)
    {
      if (strncmp(pobj->dsc, "/Page T", strlen("/Page T")) == 0)
      {
        sprintf(xbuf, "%d 0 R ", pobj->no);
        if ((strlen(xbuf) + strlen(kids)) < lkids)
          strcat(kids, xbuf);
      }
      pobj = pobj->next;
    }
    pobj = this->GetObjectStart();
    while (pobj != NULL)
    {
      if (strncmp(pobj->dsc, "/Page I", strlen("/Page I")) == 0)
      {
        sprintf(xbuf, "%d 0 R ", pobj->no);
        if ((strlen(xbuf) + strlen(kids)) < lkids)
          strcat(kids, xbuf);
      }
      pobj = pobj->next;
    }
  }
  strcat(kids, "]");
  hbuf = new char[200 + strlen(kids)];
  if (hbuf == NULL)
  {
    put_trace("-- Mem.Allocation Error for /Kids 2 --\n");
    global_error++;
    return -1;
  }
  sprintf(hbuf, "/Type /Pages\n/Count %d\n%s\n",
                this->GetPageCount(), kids);
  this->AddDict(hbuf);

  delete kids;
  delete hbuf;
  n = this->CloseObject();
  if (n < 0)
  {
    put_trace("Write Error\n");
    global_error++;
    return -1;
  }
  return no;
}   // end PDF::WritePagesObject()

// Write a complete /Page object
// Assumes that the objects belonging to a page have been
// registered via RegisterPageObject()
// Combines StartPage() and ClosePage()
// Returns: the number of the /Page object written
int PDF::MakePage(int pages_obj,
                  char *subtype)   // " T" for text, " I" for image
{
int page_obj;
char hbuf[128];
  if (logging)
  {
    put_trace("PDF::MakePage()\n");
  }
  strcpy(hbuf, "/Page");
  strcat(hbuf, subtype);
  page_obj= this->StartObject(hbuf, 1);
  this->StartDict();
  sprintf(hbuf, "/Type /Page\n/Parent %d 0 R\n", pages_obj);
  this->AddLine(hbuf);
  sprintf(hbuf, "/MediaBox [0 0 %d %d]\n", 
          this->page_xdim, this->page_ydim);
  this->AddLine(hbuf);
  sprintf(hbuf, "/Rotate %d\n", this->rotation);
  this->AddLine(hbuf);
  //
  this->WritePageResources();
  //
  this->DeletePageObjects();
  //
  return page_obj;
}   // end PDF::MakePage()

// destroy the collected page objects (local page chain ,not global chain!)
void PDF::DeletePageObjects()
{
PDFOBJ *pobj;
PDFOBJ *pn;
  pobj = this->p_first_obj;
  while (pobj != NULL)
  {
    pn = pobj->next;
    delete pobj;
    pobj = pn;
  }
  this->p_first_obj = NULL;
  this->p_last_obj  = NULL;
}   // end PDF::DeletePageObjects()

// destroy the collected global objects
void PDF::DeleteGlobalObjects()
{
PDFOBJ *pobj;
PDFOBJ *pn;
  pobj = this->g_first_obj;
  while (pobj != NULL)
  {
    pn = pobj->next;
    delete pobj;
    pobj = pn;
  }
  this->g_first_obj = NULL;
  this->g_last_obj  = NULL;
}   // end PDF::DeleteGlobalObjects()

// Add an object to the local page chain
int PDF::RegisterPageObject(int objno, char *ptype)
{
char hbuf[128];
PDFOBJ *pobj;
  if (logging)
  {
    sprintf(hbuf, "PDF::RegisterPageObject(): %d %s\n",
            objno, ptype);
    put_trace(hbuf);
  }
  pobj = new PDFOBJ(objno, 0, 0L, ptype);
  if (pobj == NULL)
    return -1;
  if (p_first_obj == NULL)
  {
    p_first_obj = pobj;
    p_last_obj  = NULL;
  }
  else
    p_last_obj->next = pobj;
  p_last_obj = pobj;
  return 0;
}   // end PDF::RegisterPageObject()

// Called while writing a /Page object
// Writes the page resources: the fonts used, the annotation objects
// (the links) and the /Content object array
int PDF::WritePageResources()
{
int  fonts, procset, objn;
char xbuf[40];
char hbuf[256];
PDFOBJ *pobj, *pn;
  if (logging)
  {
    put_trace("PDF::WritePageResources()\n");
  }
  sprintf(hbuf, "/Resources\n");
  this->AddLine(hbuf);
  this->StartDict();   // start the resources dictionary
  //
  // keep notice of the fonts used (global fonts, not local on page!)
  pobj = this->GetObjectStart();
  strcpy(hbuf, "/Font <<\n");
  fonts = 0;
  while (pobj != NULL)
  {
    if (strcmp(pobj->dsc, "/Font") == 0)
    {
      fonts++;
      sprintf(xbuf, "/F%d %d 0 R ", pobj->various, pobj->no);
      strcat(hbuf, xbuf);
    }
    pobj = pobj->next;
  }
  if (fonts > 0)
  {
    strcat(hbuf, " >>\n");
    this->AddLine(hbuf);
  }
  //
  // Write a pointer to the global ProcSet object
  procset =  this->GetObjectByType("Procset", 0);
  if (procset != 0)
  {
    sprintf(hbuf, "/ProcSet %d 0 R\n", procset);
    this->AddLine(hbuf);
  }
  //
  this->CloseDict();  // close the resources dictionary
  //
  // keep notice of the content objects on the page
  if (logging)
  {
    sprintf(hbuf, "\tPDF::ClosePage(): einfuegen /Contents\n");
    put_trace(hbuf);
  }
  strcpy(hbuf, "/Contents [");
  pobj = this->p_first_obj;
  while (pobj != NULL)
  {
    if (logging)
    {
      sprintf(xbuf, "\tpobj: %d %s\n", pobj->no, pobj->dsc);
      put_trace(xbuf);
    }
    if (strcmp(pobj->dsc, "Text") == 0)
    {
      sprintf(xbuf, "%d 0 R ", pobj->no);
      if ((strlen(hbuf) + strlen(xbuf)) < (sizeof(hbuf)-2))
        strcat(hbuf, xbuf);
    }
    pobj = pobj->next;
  }
  strcat(hbuf, "]\n");  
  this->AddLine(hbuf);
  //
  // keep notice of the link objects
  strcpy(hbuf, "/Annots [");
  pobj = this->p_first_obj;
  objn = 0;
  while (pobj != NULL)
  {
    if (strcmp(pobj->dsc, "Link") == 0)
    {
      objn++;
      sprintf(xbuf, "%d 0 R ", pobj->no);
      if ((strlen(hbuf) + strlen(xbuf)) < (sizeof(hbuf)-2))
        strcat(hbuf, xbuf);
    }
    pobj = pobj->next;
  }
  strcat(hbuf, "]\n");  
  if (objn > 0)
    this->AddLine(hbuf);
  //
  this->CloseDict();
  this->CloseObject();
  this->DeletePageObjects();
  return 0;
}   // end PDF::WritePageResources()

// Must be followed by ClosePage()
// RegisterPageObject() should be used to register the objects
// on the page (used to generate the /Kids array)
// Returns: the number of the object created
int PDF::StartPage(int pages_obj,
                   char *subtype)   // " T" for text, " I" for image
{
int page_obj;
char hbuf[128];
  if (logging)
  {
    sprintf(hbuf, "PDF::StartPage()\n");
    put_trace(hbuf);
  }
  strcpy(hbuf, "/Page");
  strcat(hbuf, subtype);
  page_obj= this->StartObject(hbuf, 1);
  this->StartDict();
  sprintf(hbuf, "/Type /Page\n/Parent %d 0 R\n", pages_obj);
  this->AddLine(hbuf);
  sprintf(hbuf, "/MediaBox [0 0 %d %d]\n", 
          this->page_xdim, this->page_ydim);
  this->AddLine(hbuf);
  sprintf(hbuf, "/Rotate %d\n", rotation);
  this->AddLine(hbuf);
  //
  // initialize the local objects chain for this page
  this->DeletePageObjects();
  return page_obj;
}    // end StartPage()

// Generates the /Resources dictionary 
// Generates the /Contents array for the actual page
// Generates the /Annots array for the actual page
// Uses heavily the local page objects chain
// Destroys the local page objects chain
int PDF::ClosePage()
{
  if (logging)
  {
    put_trace("ClosePage() 1\n");
  }
  this->WritePageResources();
  this->DeletePageObjects();
  return 0;
}   // end PDF::ClosePage()

// Generates the /Resources dictionary 
// Generates the /Contents array for the actual page
// Generates the /Annots array for the actual page
// Uses heavily the local page objects chain
// Destroys the local page objects chain
int PDF::ClosePage(int pages_obj,
                   char *subtype)
{
int  page_obj;
char hbuf[256];
PDFOBJ *pobj, *pn;
  if (logging)
  {
    put_trace("PDF::ClosePage() 2\n");
  }
  strcpy(hbuf, "/Page");
  strcat(hbuf, subtype);
  page_obj= this->StartObject(hbuf, 1);
  this->StartDict();
  sprintf(hbuf, "/Type /Page\n/Parent %d 0 R\n", pages_obj);
  this->AddLine(hbuf);
  sprintf(hbuf, "/MediaBox [0 0 %d %d]\n", 
          this->page_xdim, this->page_ydim);
  this->AddLine(hbuf);
  sprintf(hbuf, "/Rotate %d\n", rotation);
  this->AddLine(hbuf);
  this->WritePageResources();
  this->DeletePageObjects();
  return 0;
}   // end PDF::ClosePage()

// Return the current date in format TT.MM.JJJJ
char *PDF::GetDate()
{
time_t ltime;
struct tm *newtime;
short  offset;
static char dest[32];
  time(&ltime);
  newtime = localtime(&ltime);
  offset  = 1900;
  sprintf(dest,"%02d.%02d.%04d",
          (int)newtime->tm_mday,
          (int)newtime->tm_mon+1,
          (int)newtime->tm_year + offset);
  return dest;
}   // end PDF::GetDate()

// Returns: the start pos. of the required token (number of token given)
int GetTokenPos(char *buf, int tokno)
{
int pos, n, tkn;
  tkn  = 0;
  pos  = -1;
  for (n=0; buf[n] != 0; n++)
  {
    if (buf[n] == '>') break;
    if (buf[n] == '<') continue;
    if (buf[n] == ' ')
    {
      tkn++;
      if (tkn > tokno) break;
      continue;
    }
    if (tkn == tokno)
    {
      pos = n;
      break;
    }
  }
  return pos;  
}   // end GetTokenPos()

// Prepares a line of text to be entered in a PDF document
// Creates Escapes for critical characters
// Returns: the modified line of text
// !!! The returned buffer is allocated here and must be freed outside!!!
// Side effects: this->actionflag is set when a link is found
//               this->pagebreak is set when a page break tag is detected
//               this->paragraph is set when a paragraph tag is detected
//               this->font_change is set when a font change is detected
char *PDF::PrepareTextLine(char *txt)
{
int  ll, n, m;
char *tbuf1;
char *caddr;
char hbuf[512];
  if (logging)
  {
    sprintf(hbuf,"PrepareTextLine():%s\n",txt);
    put_trace(hbuf);
  }
  this->pagebreak  = 0;
  this->paragraph  = 0;
  this->actionflag = 0;
  this->font_change= 0;
  //
  // Escape Sequenzen fuer '(' und ')' erzeugen
  ll = (2 * strlen(txt)) + 1;
  tbuf1 = new char[ll];
  if (tbuf1 == NULL)
    return NULL;
  tbuf1[0] = 0;
  if (strstr(txt, "<PAGE>") != NULL)
  {
    this->pagebreak = 1;
    return tbuf1;
  }
  if ((caddr = strstr(txt, "<NEED ")) != NULL)
  {
    // check if we have a minimum of n lines
    // if not, signal a page break
    n = GetTokenPos(caddr, 1);
    if (n >= 0)
    {
      m = atoi(caddr + n) * this->leading;
      if (this->current_ypos < (m + (4 * this->leading)))
        this->pagebreak = 1;
      //sprintf(hbuf, "n=%d m=%d\n", n, m);
      //put_trace(hbuf);
    }
    return tbuf1;
  }
  if (strstr(txt, "<P>") != NULL)
  {
    this->paragraph = 1;
    return tbuf1;
  }
  if ((caddr = strstr(txt, "<FONT ")) != NULL)
  {
    // Parameter 1 = font number
    this->font_change = 1;
    n = GetTokenPos(caddr, 1);
    if (n >= 0)
      this->fontno = atoi(caddr + n);
    //
    // Parameter 2 = font size
    n = GetTokenPos(caddr, 2);
    if (n >= 0)
    {
      this->fontsize = atoi(caddr + n);
      // automatically adjust the leading
      this->current_leading = this->fontsize + 1;
    }
    //
    // Parameter 3 = leading
    n = GetTokenPos(caddr, 3);
    if (n >= 0)
    {
      this->current_leading = atoi(caddr + n);
    }
    //sprintf(hbuf, "FONT tag: %d %d %d\n", this->fontno, this->fontsize,
    //    this->current_leading);
    //put_trace(hbuf);
    return tbuf1;
  }
  if ((caddr = strstr(txt, "<EXTLINK ")) != NULL)
  {
    // link to an external file
    // put the link destination(file name) into the returned buffer
    strcpy(tbuf1, caddr + 9);
    for (n=0; tbuf1[n] != 0; n++)
      if (tbuf1[n] == '>') break;
    tbuf1[n] = 0;
    this->actionflag = 1;
    return tbuf1;
  }
  m = 0;
  for (n=0; txt[n] != 0; n++)
  {
    if (m >= (ll - 4))
      break;
    if (txt[n] == '(')
    {
      tbuf1[m++] = '\\';
      tbuf1[m++] = '(';
      continue;
    }
    if (txt[n] == ')')
    {
      tbuf1[m++] = '\\';
      tbuf1[m++] = ')';
      continue;
    }
    tbuf1[m++] = txt[n];
  }
  tbuf1[m] = 0;
  return tbuf1;
}   // end PDF::PrepareTextLine()

// Read text (containing HTML like tags) from an ASCII file
// and insert this text as the first page in the document
// Generates /Page objects for the text
// Assumes that a /Pages object has been registered!
int PDF::InsertTextFromFile(char *filename,
                            char *header1,
                            char *header2,
                            char *footer)
{
int  ll, ypos, page_number, lead;
int  xp, txt_obj, page_obj, header_obj, footer_obj;
int  procset_obj;
char *head1;
char *head2;
char *foot;
char *line_addr;
char hbuf[256];
FILE *txt_fh;
  if (logging)
    put_trace("PDF::InsertTextFromFile()\n");
  //
  txt_fh = fopen(filename, "r");
  if (txt_fh == NULL)
  {
    put_trace("Error: TOC file not found\n");
    return -1;
  }
  //
  this->rotation    = 0;
  this->ReleaseTextContent();
  //
  // Loop over the lines in the text file
  // Detect a very restricted set of tags in the file
  // (this should be revised)
  this->font_change = 0;
  lead              = this->leading;
  this->current_ypos= this->page_ydim - (lead * 7);
  page_number       = 0;
  if (logging)
    put_trace("\n");
  while (1)
  {
    if (fgets(hbuf, sizeof(hbuf), txt_fh) == NULL)
      break;
    ll = strlen(hbuf);
    if (hbuf[ll-1] == '\n')
      hbuf[ll-1] = 0;
    line_addr = this->PrepareTextLine(hbuf);
    if (line_addr == NULL)
    {
      put_trace("+++ Not Enough Memory +++\n");
      return -2;
    }
    if (this->actionflag)
    {
      int xp, xl, obj;
      //
      // link to an external file detected
      xp = this->page_xdim - this->left_margin - this->left_margin;
      xl = this->left_margin - 1;
      obj= this->MakeExtLinkAnnotation(line_addr, xp, this->current_ypos,
                                      xp+xl, this->current_ypos - (xl/2));
      this->RegisterPageObject(obj, "Link");
      delete line_addr;
      this->actionflag = 0;
      continue;
    }
    if (this->font_change)
    {
      // tag <FONT NN> detected
      lead = this->current_leading;
      delete line_addr;
      continue;
    }
    if (this->pagebreak)
    {
      // a <PAGE> tag was detected
      this->current_ypos = -1;
      delete line_addr;
      line_addr = NULL;
    }
    if (this->paragraph)
    {
      // tag <P> = start of a paragraph
      this->current_ypos -= (lead / 2);
      delete line_addr;
      continue;
    }
    //
    //
    this->current_ypos -= lead;
    //sprintf(hbuf, "lead=%d this->current_ypos=%d\n", lead, 
    //        this->current_ypos);
    //put_trace(hbuf);
    //
    //
    if (this->current_ypos < (4 * this->leading))
    {
      // page space is exhausted: finish the actual page - start a new one
      txt_obj = this->WriteTextContent();
      this->ReleaseTextContent();
      head1 = this->PrepareTextLine(header1);
      head2 = this->PrepareTextLine(header2);
      header_obj  = this->WriteHeader(head1, head2);
      foot  = this->PrepareTextLine(footer);
      page_number++;
      footer_obj  = this->WriteFooter(foot, page_number, 0);
      delete head1;
      delete head2;
      delete foot;
      procset_obj = GetObjectByType("Procset", 0);
      if (procset_obj != 0)
        this->RegisterPageObject(procset_obj, "ProcSet");
      this->RegisterPageObject(header_obj, "Text");
      this->RegisterPageObject(txt_obj, "Text");
      this->RegisterPageObject(footer_obj, "Text");
      this->SetPageCount(this->GetPageCount() + 1);
      //
      // make the current page object
      page_obj = this->MakePage(this->GetPagesObjno(), " T");
      this->current_ypos = this->page_ydim - (lead * 7);
      if (logging)
        put_trace("\n");
    }
    if (line_addr != NULL)
    {
      this->CollectTextContent(this->fontno, this->fontsize, 
                              this->left_margin, this->current_ypos, 
                              line_addr);
    }
  }  // end while 1
  //
  txt_obj = this->WriteTextContent();// text object is created and closed!
  this->ReleaseTextContent();
  procset_obj = GetObjectByType("Procset", 0);
  if (procset_obj != 0)
      this->RegisterPageObject(procset_obj, "ProcSet");
  //
  page_number++;
  head1       = this->PrepareTextLine(header1);
  head2       = this->PrepareTextLine(header2);
  header_obj  = this->WriteHeader(head1, head2);
  foot        = this->PrepareTextLine(footer);
  footer_obj  = this->WriteFooter(foot, page_number, 0);
  delete head1;
  delete head2;
  delete foot;
  this->RegisterPageObject(header_obj, "Text");
  this->RegisterPageObject(txt_obj, "Text");
  this->RegisterPageObject(footer_obj, "Text");
  this->SetPageCount(this->GetPageCount() + 1);
  fclose(txt_fh);
  //
  // Create now the /Page object for the objects on the last text page
  this->MakePage(this->GetPagesObjno(), " T");
  return 0;
}   // end PDF::InsertTextFromFile()

void  usage()
{
     fprintf(stderr,
     "Copyright 6-96 V1.5 Dr.E.Huckert, Goldbach Germany\n");
     fprintf(stderr, 
"usage: txtpdf header1 header2 footer txt-file pdf-file [-land] [-trace]\n");
}

int main(int argc, char *argv[])
{
PDF  *pdf;
int  n;
int  rotation, result;
  if (argc < 6)
  {
     usage();
     exit(1);
  }
  for (n=1; n < argc; n++)             
  {
    if (strcmp(argv[n], "-land") == 0)
      // Landscape mode: rotate by 90 degrees
      rotation = 90;
    if (strcmp(argv[n], "-trace") == 0)
      // switch on logging
      logging = 1;
  }

  //
  // Invoke constructor: opens the output file, creates
  // the essential objects, initializes the class member variables
  pdf = new PDF(argv[5], "Program txtpdf by E.Huckert");
  if (pdf == NULL)
  {
    fprintf(stderr, "--- Allocation Error ---\n");
    put_trace("Error: allocation problem\n");
    exit(3);
  }
  if (pdf->GetGlobalError() > 0)
  {
    put_trace("Global Error Found!\n");
    exit(4);
  }

  //
  // call the text to PDF conversion routine
  result = pdf->InsertTextFromFile(argv[4], 
                                   argv[1], 
                                   argv[2], 
                                   argv[3]);
  if (result < 0)
  {
    put_trace("Error: the PDF file is probable incorrect\n");
    exit(5);
  }
  if (pdf->GetGlobalError() > 0)
  {
    put_trace("Global Error Found!\n");
    exit(6);
  }

  // invoke the destructor
  // Generates the trailer with the  cross. ref.
  // Writes the page tree in the /Pages object
  delete pdf;
  //
  return 0;
}   /* end main */
